
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>
#include <errno.h>

#include "initng_global.h"

#include "initng_active_db.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "initng_execute.h"
#include "initng_common.h"
#include "initng_load_module.h"

#include "initng_handler.h"
#include "initng_kill_handler.h"

#include "initng_static_data_id.h"

/* function defines */
static void handle_WAITING_FOR_START_DEP(active_h * service_to_start);
static void handle_START_DEP_MET(active_h * service_to_start);
static void handle_START_LAUNCHED(active_h * service_to_start);


static void handle_WAITING_FOR_STOP_DEP(active_h * service_to_stop);
static void handle_STOP_DEP_MET(active_h * service_to_stop);
static void handle_STOPPING(active_h * service_to_stop);
static void handle_STOP_LAUNCHED(active_h * service_to_stop);
static void handle_STOPPED(active_h * service_to_stop);


/*
 * This funtion is run from main when g.interrupt is set.
 * Will do special actions, for some status mode.
 */
void try_start_wait_deps(void)
{
    active_h *current, *q = NULL;

    S_;

    /* walk through all active services */
    while_active_db_safe(current, q)
    {
        assert(current->name);

        /* Look at status */
        switch (current->a_status)
        {
	    case STOP_MARKED:
		mark_service(current, WAITING_FOR_STOP_DEP);
		break;
            case WAITING_FOR_STOP_DEP:
		handle_WAITING_FOR_STOP_DEP(current);
                break;
	    case STOP_DEP_MET:
		handle_STOP_DEP_MET(current);
		break;
	    case STOPPING:
		handle_STOPPING(current);
		break;
	    case STOP_LAUNCHED:
		handle_STOP_LAUNCHED(current);
		break;
	    case STOPPED:
		handle_STOPPED(current);
		break;
	    case LOADING:
		start_service(current);
		break;
            case WAITING_FOR_START_DEP:
                handle_WAITING_FOR_START_DEP(current);
                break;
	    case START_LAUNCHED:
		handle_START_LAUNCHED(current);
		break;
	    case START_DEP_MET:
		handle_START_DEP_MET(current);
            default:
                break;
        }
    }
}

/* this function will send term signal to a service if it was marked 30 seconds or more */
void term_stopping_services(void)
{
    active_h *to_term, *q = NULL;

    S_;

    while_active_db_safe(to_term, q)
    {
        if ((to_term->a_status == STOPPING || to_term->a_status == STOP_LAUNCHED ||
             to_term->a_status == STOP_LAUNCHING))
        {
            D_("Service %s is STOPPING\n", to_term->name);
            if (to_term->start_process && to_term->start_process->pid)
            {
                D_("Service %s has a start_process with a valid pid file\n",
                   to_term->name);
                if ((time(0) - to_term->time_got_status) >=
                    SECONDS_BEFORE_KILL)
                {
                    D_("Service %s has been STOPPING for %i, and its more then %i seconds\n", to_term->name, (time(0) - to_term->time_got_status), SECONDS_BEFORE_KILL);
                    /* todo, if kill(process,0) = dead then handle_killed_by_pid without warning and timeout. */

                    /* check if the process still exits! */
                    if (kill(to_term->start_process->pid, 0) < 0
                        && (errno == ESRCH))
                    {
                        /* This process is already dead */
                        F_("process %s pid %i is already dead, trying to fix.\n", to_term->name, to_term->start_process->pid);

			process_db_free(to_term->start_process);
			to_term->start_process=NULL;
			mark_service(to_term, STOPPED);
                        /*handle_killed_by_pid(to_term->start_process->pid, 1);*/
                    }
                    else
                    {

                        W_("Warning, terminating %s on pid %i!\n",
                           to_term->name, to_term->start_process->pid);
                        /* Uhm, this doesnt work : kill(-to_term->start_process->pid, SIGKILL); */
                        kill(to_term->start_process->pid, SIGKILL);
                        /* sigchld might not trigger here, for example if we're using an invalid pidfile (pid = not ours?) */
                        /*                handle_killed_by_pid(to_term->start_process->pid, 1); */
                    }
                }
                else
                {
                    D_("Time is not out ( only %i seconds ) before start killing, waiting %i more seconds.\n", (time(0) - to_term->time_got_status), SECONDS_BEFORE_KILL);
                    alarm(SECONDS_BEFORE_KILL);
                }
            }
        }
    }
}

    /* M A R K    S T U F F */

    /* When this function is called, the service is marked for starting */
int start_service(active_h * service_to_start)
{
    assert(service_to_start);
    assert(service_to_start->name);

    S_;

    if (strncmp(service_to_start->name, "virtual/", 8) == 0)
    {
        F_("Cant start a virtual service, it has to be provided!\n");
        return (FALSE);
    }

    /* check system state, if we can launch. */
    if (g.sys_state != STATE_STARTING && g.sys_state != STATE_UP)
    {
        F_("Cant start service, when system status is: %i !\n", g.sys_state);
        return (FALSE);
    }

    /* check if its alredy starting */
    if (service_to_start->a_status == WAITING_FOR_START_DEP)
    {
        D_("service %s is starting allredy.\n", service_to_start->name);
        return (TRUE);
    }
    
    /* it might actually already be up */
    if (service_to_start->a_status == DONE ||
        service_to_start->a_status == RUNNING)
    {
        D_("service %s is alredy up!\n", service_to_start->name);
        return (TRUE);
    }

    /* if service has not started stopping yet, it can be set to up directly */
    if (service_to_start->a_status == WAITING_FOR_STOP_DEP ||
        service_to_start->a_status == STOP_DEP_MET)
    {
	if(service_to_start->type==TYPE_DAEMON)
	    mark_service(service_to_start, RUNNING);
	else
	    mark_service(service_to_start, DONE);

        D_("service %s did never have time to stop!\n",
           service_to_start->name);
        return (TRUE);
    }

    /* STEP 2, service must be in one of these statuses to start */
    if (service_to_start->a_status != LOADING &&
        service_to_start->a_status != STOPPED)
    {
        W_("cant set a service with status %i:%s, to start\n",
           service_to_start->a_status,
           active_db_get_status_string(service_to_start->a_status));
        return (FALSE);
    }

    /* mark it WAITING_FOR_START_DEP and wait */
    if (!mark_service(service_to_start, WAITING_FOR_START_DEP))
    {
        W_("mark_service WAITING_FOR_START_DEP failed for service %s\n",
           service_to_start->name);
        return (FALSE);
    }
    return (TRUE);
}

    /* S T O P     S E R V I C E */
int stop_service(active_h * service_to_stop)
{
    assert(service_to_stop);

    D_("stop_service(%s);\n", service_to_stop->name);

    switch (service_to_stop->a_status)
    {
	/* IF service is stopping, do nothing. */
	case STOPPED:
	case STOP_MARKED:
	case STOP_DEP_MET:
	case WAITING_FOR_STOP_DEP:
    	    return (TRUE);
	    break;
	    
	/* this is as it shud */
	case RUNNING:
	case DONE:
	    break;
	
	/* do a warning if there is another unusual status */
	default:
	    W_("Warning, setting service %s, status %s, to STOP_MARKED!\n",
		service_to_stop->name,
		active_db_get_status_string(service_to_stop->a_status));
	    break;
    }

    /* set stopping */
    if (!mark_service(service_to_stop, STOP_MARKED))
        return (FALSE);

    /* stop all services depending on me */
    {
        active_h **deps = NULL;
        int i = 0;

        deps = active_db_dep_on_me(service_to_stop);
        while (deps[i])
        {
    	    D_("%s have to stop %s.\n", service_to_stop->name, deps[i]->name);
    	    stop_service(deps[i]);
    	    i++;
        }
        free(deps);
    }

    return (TRUE);
}

int restart_service(active_h * service_to_restart)
{
    S_;
    active_db_set(&RESTARTING, service_to_restart);
    return (stop_service(service_to_restart));
}



/* Load new process named */
active_h *start_new_service_named(const char *service)
{
    active_h *to_load = NULL;

    D_("  --  %s\n", service);

    assert(service);

    if (strncmp(service, "virtual/", 8) == 0)
    {
        W_("Virtuals can not be started manually, they have to be provided.\n");
        return (NULL);
    }

    if (!(to_load = load_to_active(service)))
    {
        /* the function calling this function will print out an error */
        F_("Unable to load active for service %s\n", service);
        return (NULL);
    }

    /* STEP 3, mark service loading, if new */
    if ((to_load->a_status == NONE) && !mark_service(to_load, LOADING))
    {
        D_("Mark service failed!\n");
        return (NULL);
    }

    return (to_load);
}

/* enter a new runlevel */
int stop_all(void)
{
    active_h *ser, *q = NULL;

    S_;
    set_sys_state(STATE_STOPPING);

    while_active_db_safe(ser, q)
    {
	/* stop services */
        stop_service(ser);
    }
    return (TRUE);
}





    /* D O    S T U F F */


/* This is the function that really launches the start-marked services if all deps are up */
static void handle_WAITING_FOR_START_DEP(active_h * service_to_start)
{
    assert(service_to_start);

    /* a check */
    if (service_to_start->a_status!=WAITING_FOR_START_DEP)
    {
	F_("Service must have status WAITING_FOR_START_DEP\n");
	return;
    }

    if (g.sys_state != STATE_STARTING && g.sys_state != STATE_UP)
    {
        F_("Cant start service, when system status is: %i !\n", g.sys_state);
        return;
    }

    /* set status to START_DEP_MET */
    mark_service(service_to_start, START_DEP_MET);
}

static void handle_START_DEP_MET(active_h * service_to_start)
{
    assert(service_to_start);

    /* a check */
    if (service_to_start->a_status!=START_DEP_MET)
    {
	F_("Service must have status START_DEP_MET\n");
	return;
    }

    if (g.sys_state == STATE_STOPPING)
    {
	mark_service(service_to_start, STOP_MARKED);
        return;
    }

    /* check if it has a start process */
    if (service_to_start->start_process)
    {
        F_("  --  (%s): Function already has a start process! Aborting\n",
           service_to_start->name);
        return;
    }

    /* F I N A L L Y   S T A R T */

    switch (start_launch(service_to_start))
    {
        case FALSE:
            D_("Done direct!\n");
            /* D O N E - D I R E C T */
            mark_service(service_to_start, DONE);
    	    return;
        case FAIL:
            F_("Did not find START file!\n");
            mark_service(service_to_start, FAIL_STARTING);
            return;                 /* return with status */
	default:
	    break;

    }

    /* make sure there are a start process */
    if (!service_to_start->start_process)
    {
        mark_service(service_to_start, FAIL_STARTING);
        return;                     /* return with status */
    }
    
    D_("Launch Started.\n");
    if(!mark_service(service_to_start, START_LAUNCHED))
    {
        mark_service(service_to_start, FAIL_STARTING);
        return;                     /* return with status */
    }
}

static void handle_START_LAUNCHED(active_h * service_to_start)
{
    assert(service_to_start);

    /* a check */
    if (service_to_start->a_status!=START_LAUNCHED)
    {
	F_("Service must have status START_LAUNCHED\n");
	return;
    }

    switch(service_to_start->type)
    {
	case TYPE_SERVICE:
	case TYPE_RUNLEVEL:
	    mark_service(service_to_start, STARTING);
	    return;
	case TYPE_DAEMON:
	    mark_service(service_to_start, RUNNING);
	    return;
	default:
	    W_("Shud not get here!\n");
	    break;
    }
}

/*
 * This function is called on interrupt for all services with
 * active status WAITING_FOR_STOP_DEP.
 */
static void handle_WAITING_FOR_STOP_DEP(active_h * service_to_stop)
{
    int i = 0;
    active_h **deps = NULL;

    assert(service_to_stop);
    
    /* a check */
    if (service_to_stop->a_status!=WAITING_FOR_STOP_DEP)
    {
	F_("Service must have status WAITING_FOR_STOP_DEP\n");
	return;
    }



    /*
     *    Check so all deps, that needs service_to_stop, is down.
     *    if there are services depending on this one still running, return false and still try
     */
    deps = active_db_dep_on_me(service_to_stop);
    while (deps[i])
    {
        D_("%s depending on %s still (status %s), waiting a bit more\n", deps[i]->name, service_to_stop->name, active_db_get_status_string(deps[i]->a_status));
	free(deps);
	return;
        i++;
    }
    free(deps);

    /* ok, stopping deps are met */
    mark_service(service_to_stop, STOP_DEP_MET);
}

static void handle_STOP_DEP_MET(active_h * service_to_stop)
{
    /* a check */
    if (service_to_stop->a_status!=STOP_DEP_MET)
    {
	F_("Service must have status STOP_DEP_MET\n");
	return;
    }

    /* check if it has a stop process */
    if (service_to_stop->stop_process)
    {
        F_("  --  (%s): Function already have a stop process! Aborting!\n",
           service_to_stop->name);
        mark_service(service_to_stop, FAIL_STOPPING);
        return;
    }

    /* F I N A L L Y   S T O P */
    mark_service(service_to_stop, STOPPING);
}

static void handle_STOPPING(active_h * service_to_stop)
{
    /* a check */
    if (service_to_stop->a_status!=STOPPING)
    {
	F_("Service must have status STOPPING\n");
	return;
    }

    /* launch stop service */
    if (stop_launch(service_to_stop) == FAIL)
    {
        F_("  --  (%s): fail launch stop!\n", service_to_stop->name);
        mark_service(service_to_stop, FAIL_STOPPING);
        return;
    }

    mark_service(service_to_stop, STOP_LAUNCHED);
}

static void handle_STOP_LAUNCHED(active_h * service_to_stop)
{
    /* a check */
    if (service_to_stop->a_status!=STOP_LAUNCHED)
    {
	F_("Service must have status STOP_LAUNCHED\n");
	return;
    }


    /* if we got a stop process with a pid */
    if (service_to_stop->stop_process && service_to_stop->stop_process->pid)
    {
        alarm(SECONDS_BEFORE_KILL + 1);     /* the alarm will try to kill stopping services */
        D_("Clock is set, %s stop script, have %i seconds to kill the service.\n", service_to_stop->name, SECONDS_BEFORE_KILL);
        return;                      /* return from here */
    }

    /* T R Y 2 , Kill start_process directly */
    if (service_to_stop->start_process
        && service_to_stop->start_process->pid > 0)
    {
    
	if(kill(service_to_stop->start_process->pid, 0) && errno == ESRCH)
	{
	    F_("Trying to kill a service (%s) with a pid (%d), but there exists no process with this pid!\n", service_to_stop->name, service_to_stop->start_process->pid); 
	    process_db_free(service_to_stop->start_process);
	    service_to_stop->start_process=NULL;
            /*handle_killed_by_pid(service_to_stop->start_process->pid, 1);*/
	} else {    	    
    	    D_("  --  (%s): REAL KILLING PID %i!\n",
        	service_to_stop->name, service_to_stop->start_process->pid);

    	    /* Uhm, this doesnt work : kill(-service_to_stop->start_process->pid, SIGTERM); */
    	    kill(service_to_stop->start_process->pid, SIGTERM);

    	    alarm(SECONDS_BEFORE_KILL + 1);     /* the alarm will try to kill stopping services */
    	    D_("Clock is set, killed process %i (%s), have %i seconds to quit before SIGKILL\n", service_to_stop->start_process->pid, service_to_stop->name, SECONDS_BEFORE_KILL);
    	    /*kill(service_to_stop->start_process->pid,SIGKILL); */
    	    return;
	}
    }
    /* if this is a empty dep service we show up here */


    /* remember this in history */
    mark_service(service_to_stop, STOPPED);
}


static void handle_STOPPED(active_h * service_to_stop)
{
    int i;
    active_h **deps = NULL;
    /* a check */
    if (service_to_stop->a_status!=STOPPED)
    {
	F_("Service must have status STOPPED\n");
	return;
    }

    /*
     * Make sure there is no services that needs this
     * that still think its running.
     */
    deps = active_db_dep_on_me(service_to_stop);
    i=0;
    while (deps[i])
    {
	if(deps[i]->type!=TYPE_RUNLEVEL)
	{
	    D_("%s have to stop %s.\n", service_to_stop->name, deps[i]->name);
	    stop_service(deps[i]);
	}
	i++;
    }
    free(deps);


    /* check if this service is restarting */
    if (active_db_is(&RESTARTING, service_to_stop))
    {
	active_db_remove(&RESTARTING, service_to_stop);
        start_service(service_to_stop);
        D_("Service is restarting now!\n");
        return;
    }


    /* free service, and forget */
    active_db_del(service_to_stop);
    active_db_free(service_to_stop);
    D_("Service removed.\n");
}
